
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "../../src/initng_global.h"
#include "../../src/process_db.h"
#include "../../src/service_db.h"
#include "../../src/initng_handler.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"


#include "initng_stcmd.h"
#include "print_service.h"


static int cmd_get_pid_of(char *arg);
static int cmd_start(char *arg);
static int cmd_start_on_new(char *arg);
static int cmd_free_service(char *arg);
static int cmd_stop(char *arg);
static int cmd_restart(char *arg);
static char *cmd_history(char *arg);
static char *cmd_active_list(char *arg);
static int cmd_print_uptime(char *arg);
static int cmd_reload(char *arg);
static int cmd_initng_restart(char *arg);
static int cmd_initng_quit(char *arg);
static char *cmd_print_modules(char *arg);
static int cmd_load_module(char *arg);
static int cmd_unload_module(char *arg);
static char *cmd_print_service_db(char *arg);
static int cmd_percent_done(char *arg);
static int cmd_set_pid(char *arg);
static int cmd_stop_unneeded(char *arg);
static char *cmd_get_father_of(char *arg);

#ifdef DEBUG
static int cmd_toggle_verbose(char *arg);
static int cmd_add_verbose(char *arg);
static int cmd_del_verbose(char *arg);
#endif


s_command GET_PID_OF =
    { 'g', INT_COMMAND, &cmd_get_pid_of, NULL, "Get pid of service" };
s_command START_SERVICE =
    { 'u', TRUE_OR_FALSE_COMMAND, &cmd_start, NULL, "Start service" };
s_command START_ON_NEW =
    { 'j', TRUE_OR_FALSE_COMMAND, &cmd_start_on_new, NULL,
    "Stop all services, and start from"
};
s_command FREE_SERVICE =
    { 'z', TRUE_OR_FALSE_COMMAND, &cmd_free_service, NULL, "Free service" };
s_command STOP_SERVICE =
    { 'd', TRUE_OR_FALSE_COMMAND, &cmd_stop, NULL, "Stop service" };
s_command RESTART_SERVICE =
    { 'r', TRUE_OR_FALSE_COMMAND, &cmd_restart, NULL, "Restart service" };
s_command PRINT_HISTORY =
    { 'l', STRING_COMMAND, NULL, &cmd_history, "Print history" };
s_command ACTIVE_LIST =
    { 's', STRING_COMMAND, NULL, &cmd_active_list, "Print active list" };
s_command PRINT_UPTIME =
    { 't', STRING_COMMAND, &cmd_print_uptime, NULL, "Print uptime" };
s_command SERVICE_RELOAD =
    { 'R', TRUE_OR_FALSE_COMMAND, &cmd_reload, NULL, "Service reload" };
s_command INITNG_RESTART =
    { 'x', TRUE_OR_FALSE_COMMAND, &cmd_initng_restart, NULL,
    "Restarts initng"
};
s_command QUIT_INITNG =
    { 'q', TRUE_OR_FALSE_COMMAND, &cmd_initng_quit, NULL, "Quits initng" };
s_command PRINT_MODULES =
    { 'm', STRING_COMMAND, NULL, &cmd_print_modules, "Print loaded modules" };
s_command LOAD_MODULE =
    { 'o', TRUE_OR_FALSE_COMMAND, &cmd_load_module, NULL, "Load Module" };
s_command UNLOAD_MODULE =
    { 'w', TRUE_OR_FALSE_COMMAND, &cmd_unload_module, NULL, "UnLoad Module" };
s_command PRINT_SERVICE_DB =
    { 'p', STRING_COMMAND, NULL, &cmd_print_service_db, "Print service_db" };
s_command PERCENT_DONE = { 'n', INT_COMMAND, &cmd_percent_done, NULL,
    "Prints percent of system up"
};
s_command SET_PID = { 'P', TRUE_OR_FALSE_COMMAND, &cmd_set_pid, NULL,
    "Set pid to a service"
};
s_command STOP_UNNEEDED =
    { 'y', TRUE_OR_FALSE_COMMAND, &cmd_stop_unneeded, NULL,
    "Stop all services, not in runlevel"
};

s_command GET_FATHER_OF = {'f', STRING_COMMAND, NULL, &cmd_get_father_of, "Print father to" };
#ifdef DEBUG
s_command TOGGLE_VERBOSE =
    { 'v', TRUE_OR_FALSE_COMMAND, &cmd_toggle_verbose, NULL,
    "Toggle the verbose flag"
};
s_command ADD_VERBOSE = { 'i', TRUE_OR_FALSE_COMMAND, &cmd_add_verbose, NULL,
    "Add string to watch for to make initng verbose"
};
s_command DEL_VERBOSE = { 'k', TRUE_OR_FALSE_COMMAND, &cmd_del_verbose, NULL,
    "Del string to watch for to make initng verbose"
};
#endif


static int cmd_get_pid_of(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (0);

    if ((apt = active_db_find_by_name(arg)) &&
        apt && apt->start_process && apt->start_process->pid)
        return (apt->start_process->pid);
    return (0);
}


static int cmd_start(char *arg)
{
    if (!arg)
        return (FALSE);

    return (start_new_service_named(arg));
}


static int cmd_start_on_new(char *arg)
{
    if (!arg)
        return (FALSE);

    g.when_out = THEN_NEW_RUNLEVEL;
    g.runlevel = i_strdup(arg);
    stop_all();
    return (TRUE);
}

static int cmd_free_service(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_by_name(arg);
    if (!apt)
        return (FALSE);

    active_db_free(apt);
    return (TRUE);
}

static int cmd_stop(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_by_name(arg);
    if (!apt)
        return (FALSE);

    return (stop_service(apt));
}

static int cmd_restart(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_by_name(arg);
    if (!apt)
        return (FALSE);

    return (restart_service(apt));
}


static char *cmd_history(char *arg)
{
    history_h *tmp = NULL;
    char *send_buf = NULL;
    char row_buf[201];
    int max = 0;

    (void) arg;

    send_buf = (char *) initng_calloc(2002, sizeof(char));

    D_("parse_control_input(): Sending History\n history:");
    while ((tmp = history_get_next(tmp)))
    {
        if (tmp->name)
        {
            D_(" %s ", tmp->name);
            sprintf(row_buf, "%i|%li|%s\n", tmp->action, (long int) tmp->time,
                    tmp->name);
            row_buf[201] = '\0';            /* to make sure it stops on buffer overflow */
            strcat(send_buf, row_buf);
        }
        if (max++ >= (HISTORY - 1))
            break;
    }
    return (send_buf);
}


static char *cmd_active_list(char *arg)
{
    active_h *apt = NULL;
    char *send_buf = NULL;
    char row_buf[201];

    (void) arg;
    D_("parse_control_input(): Sending Running process list\n sending :");

    send_buf = (char *) initng_calloc(2002, sizeof(char));

    while ((apt = active_db_next(apt)))
    {
        if (!apt->name)
            continue;
        D_(" %s->%i ", apt->name, apt->a_status);
        sprintf(row_buf, "%i|%li|%s\n", apt->a_status,
                (long int) apt->time_got_status, apt->name);
        strcat(send_buf, row_buf);
    }
    return (send_buf);
}


static int cmd_print_uptime(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);

    apt = active_db_find_by_name(arg);
    if (!apt)
        return (FALSE);

    return (difftime(time(0), apt->time_got_status));
}


static int cmd_reload(char *arg)
{
    service_h *s;
    s = service_db_find_by_name(arg);
    if (!s)
    {
        D_("Clearing complete db.\n");
        service_db_clear_all();
        return (TRUE);
    }

    D_("removing service data for %s!\n", arg);
    service_db_free(s);
    return (TRUE);
}

static int cmd_initng_restart(char *arg)
{
    (void) arg;
    g.when_out = THEN_RESTART;
    stop_all();
    return (TRUE);
}

static int cmd_initng_quit(char *arg)
{
    (void) arg;
    g.when_out = THEN_QUIT;
    stop_all();
    return (TRUE);
}



static char *cmd_print_modules(char *arg)
{
    m_h *mod = NULL;
    char *send_buf = NULL;

    (void) arg;

    send_buf = (char *) initng_calloc(2002, sizeof(char));

    strcpy(send_buf, "modules: \n");

    while ((mod = initng_get_next_module(mod)))
    {
        strcat(send_buf, mod->module_name);
        strcat(send_buf, "\n");
    }
    send_buf[2001] = '\0';
    return (send_buf);
}

static int cmd_load_module(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_load_module(arg));
}

static int cmd_unload_module(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_unload_module_named(arg));
}

static char *cmd_print_service_db(char *arg)
{
    service_h *s;
    char *data = NULL;

    if (arg && (s = service_db_find_by_name(arg)))
    {
        data = service_db_print(s);
    }
    else
    {
        data = service_db_print_all();
    }

    return (data);
}

static int cmd_percent_done(char *arg)
{
    (void) arg;

    return (active_db_percent_started());
}

static int cmd_set_pid(char *arg)
{
    active_h *apt = NULL;
    char *sname = NULL;
    int pid = 0;
    sname = initng_calloc(100, sizeof(char));
    sscanf(arg, "%i:%s", &pid, sname);
    if (!pid || !sname)
        return (FALSE);
    apt = active_db_find_by_name(sname);
    if (apt && apt->name && apt->start_process)
    {
        D_("setting pid of %s from %i to %i!\n",
           apt->name, apt->start_process->pid, pid);
        apt->start_process->pid = pid;
        if (active_db_is(&UP_WHEN_PID_SET, apt))
            apt->a_status = RUNNING;
        return (TRUE);
    }


    F_("did not found service %s to set pid %i to.", sname, pid);
    return (FALSE);
}


static int cmd_stop_unneeded(char *arg)
{
    (void) arg;
    stop_unneeded_services();
    return (TRUE);
}

static char *cmd_get_father_of(char *arg)
{
    service_h *s;
    s = service_db_find_by_name(arg);
    if (!s)
    {
	F_("Did not find service: %s\n", arg);
	return(i_strdup("Cant find service\n"));
    }
    
    if(s->father_name)
	return(i_strdup(s->father_name));

    return(i_strdup("Service has no father\n"));
}
    
#ifdef DEBUG


static int cmd_toggle_verbose(char *arg)
{
    (void) arg;
    switch (g.verbose)
    {
        case 0:
            g.verbose = 1;
            break;
        case 1:
            g.verbose = 0;
            break;
        case 2:
            g.verbose = 3;
            break;
        case 3:
            g.verbose = 2;
            break;
        default:
            g.verbose = 0;
            W_("Unknow verbose id %i\n", g.verbose);
            break;
    }
    return (g.verbose);
}

static int cmd_add_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (verbose_add(arg));
}

static int cmd_del_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (verbose_del(arg));
}
#endif

int module_init(void)
{
    D_("module_init(stcmd);\n");
    initng_command_add(&GET_PID_OF);
    initng_command_add(&START_SERVICE);
    initng_command_add(&START_ON_NEW);
    initng_command_add(&FREE_SERVICE);
    initng_command_add(&STOP_SERVICE);
    initng_command_add(&RESTART_SERVICE);
    initng_command_add(&PRINT_HISTORY);
    initng_command_add(&ACTIVE_LIST);
    initng_command_add(&PRINT_UPTIME);
    initng_command_add(&SERVICE_RELOAD);
    initng_command_add(&INITNG_RESTART);
    initng_command_add(&QUIT_INITNG);
    initng_command_add(&PRINT_MODULES);
    initng_command_add(&LOAD_MODULE);
    initng_command_add(&UNLOAD_MODULE);
    initng_command_add(&PRINT_SERVICE_DB);
    initng_command_add(&PERCENT_DONE);
    initng_command_add(&SET_PID);
    initng_command_add(&STOP_UNNEEDED);
    initng_command_add(&GET_FATHER_OF);
#ifdef DEBUG
    initng_command_add(&TOGGLE_VERBOSE);
    initng_command_add(&ADD_VERBOSE);
    initng_command_add(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(stcmd);\n");

    initng_command_del(&GET_PID_OF);
    initng_command_del(&START_SERVICE);
    initng_command_del(&START_ON_NEW);
    initng_command_del(&FREE_SERVICE);
    initng_command_del(&STOP_SERVICE);
    initng_command_del(&RESTART_SERVICE);
    initng_command_del(&PRINT_HISTORY);
    initng_command_del(&ACTIVE_LIST);
    initng_command_del(&PRINT_UPTIME);
    initng_command_del(&SERVICE_RELOAD);
    initng_command_del(&INITNG_RESTART);
    initng_command_del(&QUIT_INITNG);
    initng_command_del(&PRINT_MODULES);
    initng_command_del(&LOAD_MODULE);
    initng_command_del(&UNLOAD_MODULE);
    initng_command_del(&PRINT_SERVICE_DB);
    initng_command_del(&PERCENT_DONE);
    initng_command_del(&SET_PID);
    initng_command_del(&STOP_UNNEEDED);
    initng_command_del(&GET_FATHER_OF);
#ifdef DEBUG
    initng_command_del(&TOGGLE_VERBOSE);
    initng_command_del(&ADD_VERBOSE);
    initng_command_del(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 unloaded!\n");

}
